// openhomm - open source clone of Heroes of Might and Magic III
// Copyright (C) 2009-2010 openhomm developers team (see AUTHORS)
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
#include "precompiled.hpp"
#include "hrAdventureScreen.hpp"
#include "hrUI.hpp"

hrAdventureScreen::hrAdventureScreen()
    : dx(0), dy(0)
    , isAnimate(false)
    , isUnderground(false)
    , hasUnderground(false)
{
    connect(&scrollTimer, SIGNAL(timeout()), this, SLOT(scroll()));
    connect(&animateTimer, SIGNAL(timeout()), this, SLOT(animate()));
}

hrAdventureScreen::~hrAdventureScreen()
{
}

void hrAdventureScreen::setViewport(int width, int height)
{
    setVisibleRect(QRect(0, 0, width, height));
    window.setRect(0, 0, width, height);
    hrUI::Get().resize(width, height);
}

void hrAdventureScreen::setVisibleRect(QRect rect)
{
    rect.width() % coord::dim != 0 ?
                viewport.setWidth(rect.width() / coord::dim + 1) :
                viewport.setWidth(coord::toCell(rect.width()));
    rect.height() % coord::dim != 0 ?
                viewport.setHeight(rect.height() / coord::dim + 1) :
                viewport.setHeight(coord::toCell(rect.height()));

    /*if (!size.isEmpty() && !size.contains(viewport))
    {
        QRect rect = size.intersected(viewport);
        viewport.moveBottomRight(rect.bottomRight());
    }*/

    if (!sizeWithBorder.isEmpty() && !sizeWithBorder.contains(viewport))
    {
        QRect rect = sizeWithBorder.intersected(viewport);
        viewport.moveBottomRight(rect.bottomRight());
    }
}

void hrAdventureScreen::collect()
{
    clearItems();

    drawTiles();

    drawObjects();

    isAnimate = false;

    drawBorderTiles();
    
    QListIterator<hrGraphicsItem> it = hrUI::Get().getItems();
    while(it.hasNext())
    {
        addItem(it.next());
    }
}

void hrAdventureScreen::drawObjects()
{
    QList<hrSceneObject> &objects = isUnderground ? objectsUnderground : objectsGround;

    QMutableListIterator<hrSceneObject> it(objects);
    while (it.hasNext())
    {
        hrSceneObject &obj = it.next();

        if (viewport.intersects(obj.getRect()))
        {
            QPoint pos = coord::toPix(obj.getPoint() - viewport.topLeft());
            hrGraphicsItem item = itemsObject[obj.getId()];
            item.setCurFrame(isAnimate ? obj.getNextFrame() : obj.getCurFrame());
            item.setPoint(pos);
            addItem(item);
        }
    }

}

void hrAdventureScreen::drawTiles()
{
    const QVector<hrTile> &tiles = isUnderground ? tilesUnderground : tilesGround;

    if (tiles.isEmpty())
        return;

    for (int i = viewport.width() - 1; i >= 0; i--)
        for (int j = viewport.height() - 1; j >= 0; j--)
        {
            QPoint pos = QPoint(i, j);
            QPoint index = viewport.topLeft() + pos;
            if (!size.contains(index))
                continue;
            const hrTile &tile = tiles.at(index.y() * size.width() + index.x());
            pos = coord::toPix(pos);

            hrGraphicsItem item = itemsTerrain[tile.terrainId];
            item.setCurFrame(tile.terrainFrame);
            item.setMirror(tile.isTerrainHorizontal(), tile.isTerrainVertical());
            item.setPoint(pos);
            addItem(item);

            if (tile.hasRiver())
            {
                hrGraphicsItem item = itemsRiver[tile.riverId];
                item.setCurFrame(tile.riverFrame);
                item.setMirror(tile.isRiverHorizontal(), tile.isRiverVertical());
                item.setPoint(pos);
                addItem(item);
            }
            if (tile.hasRoad())
            {
                drawRoad(tile, pos);
            }
        }

    if (viewport.y() > 0)
    {
        for (int i = 0; i < viewport.width(); i++)
        {
            QPoint pos = QPoint(i, -1);
            QPoint index = viewport.topLeft() + pos;
            if (!size.contains(index))
                continue;
            const hrTile &tile = tiles.at(index.y() * size.width() + index.x());
            pos = coord::toPix(pos);

            if (tile.hasRoad())
            {
                drawRoad(tile, pos);
            }
        }
    }
}

void hrAdventureScreen::drawRoad(const hrTile &tile, const QPoint &pos)
{
    hrGraphicsItem item = itemsRoad[tile.roadId];
    item.setCurFrame(tile.roadFrame);
    item.setMirror(tile.isRoadHorizontal(), tile.isRoadVertical());
    item.setPoint(pos.x(), pos.y() + coord::dim / 2);
    addItem(item);
}

void hrAdventureScreen::drawBorderTiles()
{
    for (int i = 0; i < viewport.width(); i++)
        for (int j = 0; j < viewport.height(); j++)
        {
            QPoint pos = QPoint(i, j);
            QPoint index = viewport.topLeft() + pos;

            if (size.contains(index))
                continue;

            index -= sizeWithBorder.topLeft();
            const hrTile &tile = tilesBorder[index.y() * sizeWithBorder.width() + index.x()];
            pos = coord::toPix(pos);

            hrGraphicsItem item = itemsTerrain[tile.terrainId];
            item.setCurFrame(tile.terrainFrame);
            item.setPoint(pos);
            addItem(item);
        }
}

void hrAdventureScreen::loadMap(hrH3MReader *reader)
{
    Q_ASSERT(reader);

    clearMap();

    size.setSize(reader->getSize());

    makeBorder();

    hasUnderground = reader->hasUnderground();

    hrCache& cache = hrCache::Get();

    for (int i = 0; i < size.width() * size.height(); i++)
    {
        hrTile tile = reader->getTile(i);
        tilesGround.append(tile);
        loadTile(tile);
        if (hasUnderground)
        {
            tile = reader->getTile(i, true);
            tilesUnderground.append(tile);
            loadTile(tile);
        }
    }


    for (int i = 0; i < reader->getObjectsCount(); i++)
    {
        hrSceneObject object = reader->getObject(i);

        hrGraphicsItem item;
        if (!itemsObject.contains(object.getId()))
        {
            QString name = reader->getObjectName(object.getId());
            item = cache.loadItem(name);
            itemsObject[object.getId()] = item;
        }
        else
        {
            item = itemsObject[object.getId()];
        }
        object.setFrames(item.countFrames);
        object.setSize(coord::toCell(item.getSize()));

        if (!object.isUnderground())
            objectsGround.append(object);
        else
            objectsUnderground.append(object);
    }
    qSort(objectsGround);
    qSort(objectsUnderground);

    loadCursor("cradvntr.def");
    setCursor(CURSOR_POINTER);

    hrUI &ui = hrUI::Get();
    ui.addGlobalObject(this, "AdventureScreen");
    ui.script("scripts/advscr.js");
    ui.resize(width(), height());

    animateTimer.start(200);
}

void hrAdventureScreen::loadTile(const hrTile &tile)
{
    hrCache& cache = hrCache::Get();
    if (!itemsTerrain.contains(tile.terrainId))
    {
        itemsTerrain[tile.terrainId] = cache.loadItem(tile.getTerrainName(), true);
    }
    if (tile.hasRiver() && !itemsRiver.contains(tile.riverId))
    {
        itemsRiver[tile.riverId] = cache.loadItem(tile.getRiverName(), true);
    }
    if (tile.hasRoad() && !itemsRoad.contains(tile.roadId))
    {
        itemsRoad[tile.roadId] = cache.loadItem(tile.getRoadName(), true);
    }
}

void hrAdventureScreen::switchGround()
{
    if (hasUnderground)
    {
        isUnderground ? isUnderground = false : isUnderground = true;
        emit sceneChanged();
    }
}

void hrAdventureScreen::clearMap()
{
    tilesGround.clear();
    tilesUnderground.clear();
    objectsGround.clear();
    objectsUnderground.clear();
}

const hrTile& hrAdventureScreen::getTile(const QPoint &p) const
{
    Q_ASSERT(size.contains(p));

    const QVector<hrTile> &tiles = isUnderground ? tilesUnderground : tilesGround;

    return tiles.at(p.y() * size.width() + p.x());
}

int hrAdventureScreen::width() const
{
    return window.width();
}

int hrAdventureScreen::height() const
{
    return window.height();
}

void hrAdventureScreen::scroll()
{
    bool isScroll = false;
    //if (size.contains(viewport.translated(dx, dy)))
    if (sizeWithBorder.contains(viewport.translated(dx, dy)))
    {
        isScroll = true;
    }
    else
    {
        if (dx)
        {
            //if (size.contains(viewport.translated(dx, 0)))
            if (sizeWithBorder.contains(viewport.translated(dx, 0)))
            {
                dy = 0;
                isScroll = true;
            }
        }
        if (dy)
        {
            //if (size.contains(viewport.translated(0, dy)))
            if (sizeWithBorder.contains(viewport.translated(0, dy)))
            {
                dx = 0;
                isScroll = true;
            }
        }
    }

    if (isScroll)
    {
        viewport.translate(dx, dy);
        emit sceneChanged();
    }
}

void hrAdventureScreen::animate()
{
    isAnimate = true;
    emit sceneChanged();
}

void hrAdventureScreen::setCursor(AdvCursorType_e cursorType)
{
    static int oldType = -1;
    if (oldType != cursorType)
    {
        oldType = cursorType;
        emit cursorChaged(cursorType);
    }
}

void hrAdventureScreen::onMouseEvent(const QPointF &pos)
{
    //QPoint pos = event->pos();
    const int border = 50;
    bool startScrollTimer = true;

    if (pos.x() < border && pos.y() < border)
    {
        // top left
        dx = -1; dy = -1;
        setCursor(CURSOR_SCROLL_TOPLEFT);
    }
    else if (pos.x() > width() - border && pos.y() < border)
    {
        // top right
        dx = 1; dy = -1;
        setCursor(CURSOR_SCROLL_TOPRIGHT);
    }
    else if (pos.x() > width() - border && pos.y() > height() - border)
    {
        // bottom right
        dx = 1; dy = 1;
        setCursor(CURSOR_SCROLL_DOWNRIGHT);
    }
    else if (pos.x() < border && pos.y() > height() - border)
    {
        // bottom left
        dx = -1; dy = 1;
        setCursor(CURSOR_SCROLL_DOWNLEFT);
    }
    else if (pos.x() < border)
    {
        // left
        dx = -1; dy = 0;
        setCursor(CURSOR_SCROLL_LEFT);
    }
    else if (pos.x() > width() - border)
    {
        // right
        dx = 1; dy = 0;
        setCursor(CURSOR_SCROLL_RIGHT);
    }
    else if (pos.y() < border)
    {
        // up
        dx = 0; dy = -1;
        setCursor(CURSOR_SCROLL_TOP);
    }
    else if (pos.y() > height() - border)
    {
        // down
        dx = 0; dy = 1;
        setCursor(CURSOR_SCROLL_DOWN);
    }
    else
    {
        // stop
        dx = 0; dy = 0;
        startScrollTimer = false;
        setCursor(CURSOR_POINTER);
    }

    const int delay = 20;

    if (startScrollTimer)
    {
        if (!scrollTimer.isActive())
            scrollTimer.start(delay);
    }
    else
    {
        scrollTimer.stop();
    }
}

void hrAdventureScreen::onMousePressEvent(const QPoint &pos)
{
    hrUI::Get().onMousePress(pos);
    emit sceneChanged();
}

void hrAdventureScreen::onMouseReleaseEvent(const QPoint &pos)
{
    hrUI::Get().onMouseRelease(pos);
    emit sceneChanged();
}

void hrAdventureScreen::onMouseLeave()
{
    scrollTimer.stop();
}

void hrAdventureScreen::makeBorder()
{
    const int MapBorder = 10;

    sizeWithBorder.setRect(  -MapBorder
                           , -MapBorder
                           , size.width() + 2 * MapBorder
                           , size.height() + 2 * MapBorder
                           );

    hrTile tile;
    tile.terrainId = 10; // edg.def
    loadTile(tile);

    tilesBorder.clear();

    for (int i = 0; i < sizeWithBorder.width(); i++)
        for (int j = 0; j < sizeWithBorder.height(); j++)
        {
            hrTile tile;
            QPoint index = sizeWithBorder.topLeft() + QPoint(i, j);
            if (!size.contains(index))
            {
                tile.terrainId = 10; // edg.def
                if (index + QPoint(1, 1) == size.topLeft())
                {
                    tile.terrainFrame = 16;
                }
                else if (index + QPoint(-1, 1) == size.topRight())
                {
                    tile.terrainFrame = 17;
                }
                else if (index + QPoint(-1, -1) == size.bottomRight())
                {
                    tile.terrainFrame = 18;
                }
                else if (index + QPoint(1, -1) == size.bottomLeft())
                {
                   tile.terrainFrame = 19;
                }
                else if (size.contains(index + QPoint(0, 1))
                         && index.y() + 1 == size.top())
                {
                    tile.terrainFrame = 20 + qrand() % 4;
                }
                else if (size.contains(index + QPoint(1, 0))
                          && index.x() + 1 == size.left())
                {
                    tile.terrainFrame = 32 + qrand() % 4;
                }
                else if (size.contains(index + QPoint(0, -1))
                         && index.y() - 1 == size.bottom())
                {
                    tile.terrainFrame = 28 + qrand() % 4;
                }
                else if (size.contains(index + QPoint(-1, 0))
                          && index.x() - 1 == size.right())
                {
                    tile.terrainFrame = 24 + qrand() % 4;
                }
                else
                {
                    tile.terrainFrame = qrand() % 15;
                }
                tilesBorder[j * sizeWithBorder.width() + i] = tile;
            }
        }

}
